open List;;
open String;;

(*'hist' will be modified for symbols which appear 
* more than once in the same term;
* 'num_node' is the number of the node 
* in the configuration graph 
*)
type hist_node={hist:int;num_node:int};;

type lint= hist_node list;;

(* 
* a list of 'hist' elements meaning the history of 
* different symbols in a term
*)
type llint= lint ref;;

type kind= 
    Omicron
  | Arrow of kind*kind;;

type term=
    Symbol of char*llint
  | App of term*term;;

type vars= char list;;

(* terms of the form \lambda x.t*)
type lambda_term={v:vars;lambdat:term};;

(*
* will be used for mapping symbols
* (terminals and non-terminals) to kinds
*)
type pair= C of char| K of kind;;

(*
* will be used for mapping 
* non-terminals to lambda-terms
*)
type triple=S of char| V of vars| T of term;;

type pair_list=pair list;;
type triple_list=triple list;;
type kind_list=kind list;;
   
let rec arity k = match k with
    Omicron -> 0
  | Arrow(k1,k2)-> (arity k2)+1;;

let rec order k=match k with
    Omicron -> 0
  | Arrow(k1,k2)-> max ((order k1)+1) (order k2) ;;

(* 
* gives the string representation of the term 'te',
* using '_' instead of open brackets
*) 
let rec string_of_term te=
match te with 
  Symbol(c,li)->Char.escaped c
  | App(t1,t2)->"_"^(string_of_term t1)^(string_of_term t2)^"_"
;;

(* 
* gives the string representation of the term 'te',
* using the open brackets
*) 
let rec string_of_termb te=
match te with 
  Symbol(c,li)->Char.escaped c
  | App(t1,t2)->"("^(string_of_termb t1)^(string_of_termb t2)^")"
;;

(*
* gives the string representation of a list of terms;
* the strings of the terms are concatenated,
* with no delimiter between them
*)
let rec string_of_terms list_terms=
match list_terms with 
  te::li->(string_of_term te)^(string_of_terms li)
  |[]->"";;

(*
* gives the string representation of a list of terms;
* the strings of the terms are concatenated,
* with no delimiter between them;
* open brackets are showed
*)
let rec string_of_termsb list_terms=
match list_terms with 
  te::li->(string_of_termb te)^(string_of_termsb li)
  |[]->"";;

(*
* gives the string representation of a list of terms;
* the strings of the terms are concatenated,
* with the '_' delimiter between them
*)
let string_of_terms2 list_terms=
let s=ref "" in
for i=0 to (List.length list_terms -1) do
s:=!s^ (string_of_term (List.nth list_terms i))^"_";
done;
!s;;

(*
* transforms the list of integers representing the 
* history of a symbol to a string
*)
let rec string_of_list l=
match l with
  e::lrest->(string_of_int e.hist)^(string_of_list lrest)
  |[]->"";;

(*
*gives the string representation of a term
*including the history of each symbol
*)
let rec string_of_term_numbers t=
match t with
  Symbol(c,li)->if (List.length !li)>0 then 
                  (Char.escaped c) ^(string_of_list  !li ) 
               else Char.escaped c
  |App(t1,t2)->(string_of_term_numbers t1)^(string_of_term_numbers t2);;


(*
*gives the string representation of a list of terms
*including the history of each symbol
*)
let rec string_of_terms_numbers list_terms=
match list_terms with 
  te::li->(string_of_term_numbers te)^(string_of_terms_numbers li)
  |[]->"";;


(*prints a list of variables(integers) to 'myfile'*)
let rec output_vars myfile v=
match v with
  c::l->output_char myfile c; 
        output_vars myfile l
  |[]->output_char myfile '\n';;


(*prints a term to 'myfile'*)
let rec output_term myfile t= 
match t with
  Symbol(c,li)->output_char myfile c
  |App(t1,t2)->output_char myfile '('; 
               output_term myfile t1;
               output_term myfile t2;
               output_char myfile ')';;

(*prints an element of R( from a HORS) to 'myfile'*)
let rec output_triple_list myfile pl= 
match pl with
  S(c)::V(v)::T(t)::l->output_char myfile c; 
                       output_char myfile ' ';
                       output_vars myfile v; 
                       output_term myfile t; 
                       output_char myfile '\n'; 
                       output_triple_list myfile l
  |[] ->output_char myfile '\n';;

(*
* returns the list of elements of R, which is constructed from 'pl'.
* 'pl' will be provided by the parser
*)
let rec get_triple_list pl= 
match pl with
  S(c)::V(v)::T(t)::l->(c, v, t)::get_triple_list l
  |[] ->[];;

(*prints the kind 'k' to 'myfile'*)
let rec output_kind myfile k=
match k with
  Omicron ->output_string myfile " o "
 |Arrow(k1,k2)->output_char myfile '(';  
                output_kind myfile k1;
                output_kind myfile k2;
                output_char myfile ')';;

(*
* used for printing the kinds of terminals and 
* non-terminals to 'myfile'
*)
let rec output_pair_list myfile pl=
match pl with
  C(c)::K(p)::l->output_char myfile c;
                 output_kind myfile p; 
                 output_char myfile '\n'; 
                 output_pair_list myfile l
  |[] ->output_char myfile '\n';; 

(*
* returns the list of mappings from symbols to kinds.
* 'pl' will be provided by the parser
*)
let rec get_pair_list pl= 
match pl with
  C(c)::K(p)::l->(c,p)::get_pair_list l
  |[] ->[];;

(*prints a list of kinds to 'myfile'*)
let rec output_kind_list myfile pl= 
match pl with
   p::l->output_kind myfile p; 
         output_char myfile '\n';
         output_kind_list myfile l
  |[] ->output_char myfile '\n';;

(*returns a list of the symbols in term 't'*)
let rec list_of_term t=
match t with 
  Symbol(c,l)->[c]
  |App(t1,t2)-> List.append (list_of_term t1)(list_of_term t2);;

(* 
*returns a list with the number of occurences of each symbol 
*in the list of terms 'li'
*)
let number_symbols li=
let length=(List.length li)-1 in
let li1= ref [] in
for i=0 to length do
  li1:=(List.append !li1 [ref 0])
done;
for i=0 to length do
  if !(List.nth !li1 i)==0 then 
    let k=ref 1 in 
  for j=(i+1)  to length 
  do
    if (List.nth li i)==(List.nth li j) then 
    begin 
      k:=!k+1;
      (List.nth !li1 j):=!k
    end;
  done;
  if (!k>1) then (List.nth !li1 i):= 1
done;
li1;;

(*
* computes the history of each symbol in term 't',
* which is in the node 'i', and modifies 't' 
* accordingly
*)
let compute_history  t i=
let rec  term_from_list t lnum i= 
match t with
Symbol(c,l)-> 
     if (!(List.hd !lnum)>0)then 
     begin 
       let k={hist=(!(List.hd !lnum));num_node=i} in
       if (List.length !l)>0 then l:=(List.append  !l [k] )
       else l:= [ k];
       lnum:=(List.tl !lnum) 
     end
     else  
     begin           
       lnum:=(List.tl !lnum);
       if (List.length !l)==0 then 
       let k={hist=0;num_node=i} in l:=  [k]   
     end
|App(t1,t2)-> 
     term_from_list t1 lnum i;term_from_list t2 lnum i
in 
let lnu=number_symbols (list_of_term t) in
term_from_list t lnu i;
t;;





    
   
