
module M=Mygraph;;

(*Returns the string representation of an atomic type*)
let rec string_of_type mytype=
match mytype with 
  M.I(i)->string_of_int i
  |M.A(li,ty)->
let s=ref "(" in 
let len=List.length li in
if len==1 then 
  s:="("^(string_of_type(List.nth li 0))^"-"^(string_of_type ty)^")"
else 
begin
  s:="(";
  for i=0 to (len-2) 
  do
    s:=(!s)^string_of_type(List.nth li i)^"^";
  done;
  s:=(!s)^string_of_type(List.nth li (len-1))^"-"^(string_of_type ty)^")";
end;
!s;;

(*Binds a symbol(terminal, non-terminal, variable) to its type*)
module Binding=struct
type t={c:char; ty:M.atomic_type}
let compare n m = 
if (M.equaltype n.ty m.ty)&&(n.c==m.c) then 0 
else
  let s=(Char.escaped n.c)^(string_of_type n.ty) in 
  let r=(Char.escaped m.c)^(string_of_type m.ty) in
  Pervasives.compare s r
end;;


module Type=struct
type t=M.atomic_type
let compare n m = 
if (M.equaltype n m) then 0 
else
  let s=(string_of_type n) in 
  let r=(string_of_type m) in
  Pervasives.compare s r
end;;

(*set of bindings, used for representing environments*)
module SetBindings=Set.Make(Binding)

(*set of types*)
module SetTypes=Set.Make(Type)

type listtype=M.atomic_type list;;


module ListType = struct
type t = listtype;;
let compare l1 l2 =
let r=ref 0 in
let i=ref 0 in
let len=(List.length l1) in
while !i<=(len-1) 
do
  let res=Type.compare (List.nth l1 !i) (List.nth l2 !i) in
  if res ==0 then i:=!i+1
  else 
  begin 
    i:=len; 
    r:=res;
  end
done;
!r ;;
end;;

module SetListTypes = Set.Make(ListType);;


(*
* Returns the cartesian product between a set of 
* lists of atomic_types and a set of atomic_types.
*The result is a set of lists of atomic_types.
*)
let cartesian_product s1 s2 =
let f x y t = SetListTypes.add ( (List.append x [y])) t in
let g x t = SetTypes.fold (f x) s2 t in
SetListTypes.fold g s1 SetListTypes.empty;;


(*
*Returns true if the type 'mytype' does not
* contain any type variables. 'nr' is the number of states
* in the automaton. Any type variable is an integer
* bigger than 'nr'.
*)
let rec is_without_variables mytype nr=
let s=ref true in
match mytype with 
  M.I(i)->if i>nr then s:=false;!s
  |M.A(li,ty)->
let len=List.length li in 
for i=0 to (len-1) 
do
  s:=(!s) && (is_without_variables (List.nth li i) nr);
done;
s:=(!s) && (is_without_variables ty nr);
!s;;
      
(*
*Transforms the list of types 'li' to 
*an atomic type.
*)
let atomictype_of_listtypes li=
let atomic_list=List.rev (List.tl (List.rev li)) in
let elem=List.nth li ((List.length li)-1) in
M.A(atomic_list,elem);;


(*
* The function 'elim' is the procedure from the paper. 
* It returns a set of types.
* 'nr' is the number of states of the automaton. 
* Every number bigger than 'nr', in the type 'mytype', represents a type variable. 
*)
let rec elim mytype nr=
match mytype with 
  M.I(i)->
       if i>nr then  SetTypes.empty
       else 
       begin 
         let st=ref SetTypes.empty in
         st:=SetTypes.add mytype (!st);
         !st
       end
  |M.A(li,ty)->
       let len=List.length li in 
       let lnew_elim=ref [] in
       for i=0 to (len-1) 
       do
         let st=ref SetTypes.empty in
         let inter_type=(List.nth li i) in
         st:=SetTypes.union (elim inter_type nr) (!st);
         if (is_without_variables inter_type nr)==false then 
             st:=SetTypes.add (M.I(-1)) (!st);
         lnew_elim:=List.append (!lnew_elim) [!st];
       done;
       let st=elim ty nr in 
            lnew_elim:=List.append (!lnew_elim) [st];
       let l=ref SetListTypes.empty in
             l:=SetListTypes.add [] !l;
       for j=0 to (List.length !lnew_elim)-1 
       do
         l:=cartesian_product !l (List.nth !lnew_elim j); 
       done;
       let st=ref SetTypes.empty in
       let f x=  
         st:=SetTypes.add (atomictype_of_listtypes x) !st in
       SetListTypes.iter f !l;                                   
       !st;;

(*Checks if type 'mytype' is Top*)
let isT mytype=
match mytype with
  M.I(i)->if i==(-1)then true else false
  |M.A(li,ty)->false;;

(*
* Returns the type 'mytype', 
* with Top removed from it.
*)
let rec removeT mytype=
match mytype with 
  M.I(i)->mytype
  |M.A(li,ty)->
        (let l=ref [] in
         let len=List.length li in 
         for i=0 to (len-1) 
         do
           let elem=(List.nth li i) in
           if (isT elem)==false then 
             l:=List.append !l [removeT elem];
         done;
         let newty=removeT ty in
         M.A(!l,newty));;

(*
* Returns the type 'mytype',
* where the empty lists representing intersection types
* have been replaced with Top
*)
let rec replaceEmptyList mytype=
match mytype with 
  M.I(i)->mytype
  |M.A(li,ty)->
      (if (List.length li==0) 
       then let ty2=(replaceEmptyList ty) in M.A([M.I(-1)],ty2) 
       else 
       begin 
         let l=ref [] in 
         let len=List.length li in 
         for i=0 to (len-1) 
         do
           l:=List.append !l [replaceEmptyList (List.nth li i)];
         done;
         let newty=replaceEmptyList ty in
         M.A(!l,newty);
       end);;      


(*Returns the last type I(i) of the atomic type 'typ'*) 
let rec getLastType typ=
match typ with 
  M.I(i)->typ
 |M.A(li,ty)-> getLastType ty;;


(*
*Type bindings for 'variables' are added to the environment 'envir'.
*The right types are extracted from the type 'ty'. The arity for each
* variable is added to 'char_arityF'. At the first call, 'nr' has to be 0.
* 'nr' acts as a counter for the variables.
*)
let rec addVariables variables ty envir char_arityF  nr=
match ty with 
  M.A(li,myt)->(
     let len=List.length li-1 in
     let ch=List.nth variables !nr in
     for i=0 to len 
     do
       let typ=(List.nth li i) in
       envir:=SetBindings.add {Binding.c=ch;Binding.ty=typ} !envir;
       char_arityF:= (ch,M.varArity typ)::(!char_arityF);
     done;
     nr:=!nr+1;
     addVariables variables myt envir char_arityF nr)
  |M.I(i)->nr:=!nr+1;;


(*Appends to the list l the first 'nr2' intersection types of the type 'ty2'.*)
let  sets_of_type ty2 l2 nr2=
let rec loop ty l nr =
(
match ty with 
  M.A(li, myt)->
    if (!nr<nr2) then 
    begin
      let s=ref SetTypes.empty in
      for i=0 to (List.length li)-1 
      do
        s:=SetTypes.add (List.nth li i) !s;
      done ;
      l:=List.append !l [!s];
      nr:=!nr+1;
      loop myt l nr
    end
    else
      let s=ref SetTypes.empty in
    s:=SetTypes.add ty !s;
    l:=List.append !l [!s]
  |M.I(i)->  
    let s=ref SetTypes.empty in
    s:=SetTypes.add ty !s;
    l:=List.append !l [!s];
)
in let nr1=ref 0 in 
   loop ty2 l2 nr1;;


(*Checks if the list 'l' contains only the element Top. *)
let is_list_empty l=
if (List.length l)==1 then
      let elem=List.nth l 0 in
         if M.equaltype elem (M.I(-1)) then true
         else false
else false;;

(*Returns the set of types derivable for the term 't'.*)
let rec tGamma envir char_arityF t=
let ch=M.first_symbol t in 
let ar= ref 0 in
let bound=(List.length char_arityF) -1 in
for num=0 to bound 
do
  let (a,b)=(List.nth char_arityF num) in
  if (a==ch) then ar:=b
done;
if (!ar==0) then 
begin
  let t_types=ref SetTypes.empty in
  let g x=(if x.Binding.c==ch then
             t_types:=SetTypes.add (x.Binding.ty) !t_types)
  in SetBindings.iter g envir;
  !t_types;
end
else 
let t_types=ref SetTypes.empty in
let li=M.decompose t in
let lenli=List.length li  in
if lenli==0 then 
begin
  let t_types=ref SetTypes.empty in
  let g x=(if x.Binding.c==ch then
             t_types:=SetTypes.add (x.Binding.ty) !t_types)
           in SetBindings.iter g envir;
  !t_types;
end
else           
begin      
let f x= (if x.Binding.c==ch then
          begin
            let l=ref [] in sets_of_type x.Binding.ty l lenli;
            let ok=ref 1 in
            for i=0 to (lenli-1) 
            do
              let si=(List.nth !l i) in
              let emptyset=ref SetTypes.empty in
              emptyset:=SetTypes.add (M.I(-1)) !emptyset;
              if (SetTypes.equal !emptyset si)== false then
                let tG=tGamma envir char_arityF (List.nth li i) in
                if (SetTypes.subset si tG)==false 
                then ok:=0;
            done;
            if (!ok==1) then 
            begin
              let listforunion=List.nth !l lenli in
              t_types:=SetTypes.union listforunion !t_types;
            end
          end) in 
SetBindings.iter f envir;
!t_types;
end;;


(*Prints the set of bindings 'myset' to 'myfile'.*) 
let output_setbindings myfile myset=
let f x= output_char myfile (x.Binding.c);
         M.output_type myfile (x.Binding.ty);
         output_char myfile '\n' in        
SetBindings.iter f myset;;


(*Prints the set of types 'myset' to 'myfile'.*) 
let output_settypes myfile myset=
let f x= M.output_type myfile x;
         output_char myfile '\n' in        
SetTypes.iter f myset;;

